#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from collections import abc
from xml.etree import ElementTree

from networking_baremetal import common
from networking_baremetal.openconfig.vlan import vlan


class NetworkInstances(abc.Collection):
    """Top-level grouping containing a list of network instances."""

    NAMESPACE = 'http://openconfig.net/yang/network-instance'
    TAG = 'network-instances'

    def __init__(self):
        self._network_instances = list()

    def __iter__(self):
        return iter(self._network_instances)

    def __len__(self):
        return len(self._network_instances)

    def __contains__(self, item):
        return item in self._network_instances

    @property
    def network_instances(self):
        return self._network_instances

    def add(self, name: str):
        """Add network instance

        :param name: A unique name identifying the network instance
        :type: str
        :Keyword arguments: Network instance arguments
        """
        network_instance = NetworkInstance(name)
        self._network_instances.append(network_instance)
        return network_instance

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('xmlns', self.NAMESPACE)
        for instance in self.network_instances:
            elem.append(instance.to_xml_element())
        return elem


class NetworkInstance:
    """An OpenConfig description of a network_instance.

    This may be a Layer 3 forwarding construct such as a virtual
    routing and forwarding (VRF) instance, or a Layer 2 instance
    such as a virtual switch instance (VSI). Mixed Layer 2 and
    Layer 3 instances are also supported.
    """
    NAMESPACE = 'http://openconfig.net/yang/network-instance'
    TAG = 'network-instance'

    def __init__(self, name):
        self.name = name
        self._vlans = vlan.Vlans()

    @property
    def name(self):
        """A unique name identifying the network instance"""
        return self._name

    @name.setter
    def name(self, value: str):
        """A unique name identifying the network instance"""
        if not isinstance(value, str):
            raise TypeError('name must be string, got {}'.format(type(value)))
        self._name = value

    @name.deleter
    def name(self):
        self._name = None

    @property
    def vlans(self):
        """Group/List of VLANs - keyed by id"""
        return self._vlans

    @vlans.setter
    def vlans(self, value):
        if not isinstance(value, vlan.Vlans):
            raise TypeError('vlans must be Vlans, got {}'.format(type(value)))
        self._vlans = value

    @vlans.deleter
    def vlans(self):
        self._vlans = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        if self.name:
            common.txt_subelement(elem, 'name', self.name)
        if self.vlans:
            elem.append(self.vlans.to_xml_element())
        return elem



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from collections import abc
from typing import Optional
from xml.etree import ElementTree

from networking_baremetal import common
from networking_baremetal import constants
from networking_baremetal.openconfig.vlan import types


class TrunkVlans(abc.Collection):

    def __init__(self):
        self._trunk_vlans = []

    def __iter__(self):
        return iter(self._trunk_vlans)

    def __len__(self):
        return len(self._trunk_vlans)

    def __contains__(self, item):
        return item in self._trunk_vlans

    def add(self, value):
        """Add vlan or range of vlans (range: 100..200)"""
        try:
            value = int(value)
            if value not in self._trunk_vlans:
                self._trunk_vlans.append(types.VlanId(value).vlan_id)
        except ValueError:
            if value not in self._trunk_vlans:
                self._trunk_vlans.append(types.VlanRange(value).vlan_range)


class VlanSwitchedConfig:
    """Ethernet interface VLAN config

    VLAN related configuration that is part of the physical
    Ethernet interface.
    """
    NAMESPACE = 'http://openconfig.net/yang/vlan'
    PARENT = 'switched-vlan'
    TAG = 'config'

    def __init__(self,
                 operation: str = constants.NetconfEditConfigOperation.MERGE,
                 interface_mode: Optional[str] = None,
                 native_vlan: Optional[int] = None,
                 access_vlan: Optional[int] = None):

        self.operation = operation
        self._interface_mode = None
        self._native_vlan = None
        self._access_vlan = None
        self._trunk_vlans = TrunkVlans()
        if interface_mode:
            self.interface_mode = interface_mode
        if native_vlan:
            self.native_vlan = native_vlan
        if access_vlan:
            self.access_vlan = access_vlan

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value if self._operation else None

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @operation.deleter
    def operation(self):
        self._operation = None

    @property
    def interface_mode(self):
        """Get the interface to access or trunk mode for VLANs"""
        return self._interface_mode.value if self._interface_mode else None

    @interface_mode.setter
    def interface_mode(self, value):
        """Set the interface to access or trunk mode for VLANs"""
        self._interface_mode = types.VlanInterfaceMode(value)

    @interface_mode.deleter
    def interface_mode(self):
        """Delete the interface to access or trunk mode for VLANs"""
        self._interface_mode = None

    @property
    def native_vlan(self):
        """Native VLAN

         is valid for trunk mode interfaces
         """
        return self._native_vlan.vlan_id if self._native_vlan else None

    # TODO(hjensas): Only allow if interface_mode == trunk
    @native_vlan.setter
    def native_vlan(self, value: int):
        """Set native VLAN

        is valid for trunk mode interfaces
        """
        self._native_vlan = types.VlanId(value)

    @native_vlan.deleter
    def native_vlan(self):
        """Delete native VLAN"""
        self._native_vlan = None

    # TODO(hjensas): Only allow if interface_mode == access
    @property
    def access_vlan(self):
        """Access VLAN assigned to the interfaces"""
        return self._access_vlan.vlan_id if self._access_vlan else None

    @access_vlan.setter
    def access_vlan(self, value: int):
        """Set access VLAN assigned to the interfaces"""
        self._access_vlan = types.VlanId(value)

    @access_vlan.deleter
    def access_vlan(self):
        """Unset access VLAN assigned to the interfaces"""
        self._access_vlan = None

    @property
    def trunk_vlans(self):
        """Allowed VLANs may be specified for trunk mode interfaces"""
        return self._trunk_vlans

    # TODO(hjensas): Only allow if interface_mode == trunk
    @trunk_vlans.setter
    def trunk_vlans(self, value: str):
        """Set allowed VLANs may be specified for trunk mode interfaces"""
        self._trunk_vlans.add(value)

    @trunk_vlans.deleter
    def trunk_vlans(self):
        self._trunk_vlans = TrunkVlans()

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        if self.operation:
            elem.set('operation', self.operation)
        if self.interface_mode:
            common.txt_subelement(elem, 'interface-mode',
                                  self.interface_mode)
        if self.access_vlan is not None:
            common.txt_subelement(elem, 'access-vlan', str(self.access_vlan))
        if self.native_vlan is not None:
            common.txt_subelement(elem, 'native-vlan', str(self.native_vlan))
        if self.trunk_vlans is not None:
            for item in self.trunk_vlans:
                common.txt_subelement(elem, 'trunk-vlans', str(item))
        return elem


class VlanSwitchedVlan:
    """VLAN interface-specific data on Ethernet interfaces.

    Enclosing container for VLAN interface-specific
    data on Ethernet interfaces. These are for standard
    L2, switched-style VLANs.
    """
    NAMESPACE = 'http://openconfig.net/yang/vlan'
    PARENT = 'ethernet'
    TAG = 'switched-vlan'

    def __init__(self):
        self._config = VlanSwitchedConfig()

    @property
    def config(self):
        """Configuration parameters for VLANs"""
        return self._config

    @config.setter
    def config(self, value):
        if not isinstance(value, VlanSwitchedConfig):
            raise TypeError('config must be VlanSwitchedConfig, got {}'
                            .format(type(value)))
        self._config = value

    @config.deleter
    def config(self):
        self._config = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('xmlns', self.NAMESPACE)
        if self.config:
            elem.append(self.config.to_xml_element())
        return elem


class VlanConfig:
    """OpenConfig VLAN configuration"""

    NAMESPACE = 'http://openconfig.net/yang/vlan'
    PARENT = 'vlan'
    TAG = 'config'

    def __init__(self,
                 operation=constants.NetconfEditConfigOperation.MERGE,
                 vlan_id: int = None,
                 name: str = None,
                 status: str = None):
        self.operation = operation
        self._vlan_id = None
        self._name = None
        self._status = None
        if vlan_id:
            self.vlan_id = vlan_id
        if name:
            self.name = name
        if status:
            self.status = status

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @property
    def vlan_id(self):
        """The id of the VLAN"""
        return self._vlan_id.vlan_id if self._vlan_id else None

    @vlan_id.setter
    def vlan_id(self, value: int):
        self._vlan_id = types.VlanId(value)

    @vlan_id.deleter
    def vlan_id(self):
        self._vlan_id = None

    @property
    def name(self):
        """Interface VLAN name."""
        return self._name

    @name.setter
    def name(self, value: str):
        if not isinstance(value, str):
            raise TypeError('name must be string, got {}'
                            .format(type(value)))
        self._name = value

    @name.deleter
    def name(self):
        self._name = None

    @property
    def status(self):
        """Admin state of the VLAN"""
        return self._status.value if self._status else None

    @status.setter
    def status(self, value: str):
        """Admin state of the VLAN"""
        self._status = types.VlanStatus(value)

    @status.deleter
    def status(self):
        self._status = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('operation', self.operation)
        if self.vlan_id is not None:
            common.txt_subelement(elem, 'vlan-id', str(self.vlan_id))
        if self.name is not None:
            common.txt_subelement(elem, 'name', self.name)
        if self.status is not None:
            common.txt_subelement(elem, 'status', self.status)
        return elem


class Vlan:
    """Base vlan"""

    NAMESPACE = 'http://openconfig.net/yang/vlan'
    PARENT = 'vlans'
    TAG = 'vlan'

    def __init__(self, vlan_id: int,
                 operation=constants.NetconfEditConfigOperation.MERGE):
        self.operation = operation
        self.vlan_id = vlan_id
        self._config = VlanConfig(vlan_id=self.vlan_id)

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation '
                            'attribute.'.format(type(value)))

    @property
    def vlan_id(self):
        """The id of the VLAN"""
        return self._vlan_id.vlan_id

    @vlan_id.setter
    def vlan_id(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f'vlan_id must be integer, got {type(value)}')
        self._vlan_id = types.VlanId(value)

    @vlan_id.deleter
    def vlan_id(self):
        self._vlan_id = None

    @property
    def config(self):
        """Configuration parameters for VLAN"""
        return self._config

    @config.setter
    def config(self, value):
        """Configuration parameters for VLAN"""
        if not isinstance(value, VlanConfig):
            raise TypeError('config must be VlanConfig, got {}'
                            .format(type(value)))
        self._config = value
        if self.vlan_id:
            self.config.vlan_id = self.vlan_id

    @config.deleter
    def config(self):
        self._config = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        if self.vlan_id:
            common.txt_subelement(elem, 'vlan-id', str(self.vlan_id),
                                  attrib={'operation': self.operation})
        if self.config:
            elem.append(self.config.to_xml_element())
        return elem


class Vlans(abc.Collection):
    """Group/List of VLANs"""

    NAMESPACE = 'http://openconfig.net/yang/vlan'
    TAG = 'vlans'

    def __init__(self):
        # List of vlans of type OpenconfigVlan
        self._vlans = list()

    def __iter__(self):
        return iter(self._vlans)

    def __len__(self):
        return len(self._vlans)

    def __contains__(self, item):
        return item in self._vlans

    @property
    def vlans(self):
        """List of VLANs"""
        return self._vlans

    def add(self, vlan_id: int):
        """Add VLAN

        :param vlan_id: VLAN ID
        :type: int
        :Keyword arguments: VLAN configuration
        """
        vlan = Vlan(vlan_id)
        self._vlans.append(vlan)
        return vlan

    def remove(self, vlan_id: int):
        """Remove VLAN

        :param vlan_id: VLAN ID
        :type: int
        """
        vlan = Vlan(vlan_id)
        vlan.operation = constants.NetconfEditConfigOperation.REMOVE
        self._vlans.append(vlan)
        return vlan

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('xmlns', self.NAMESPACE)
        for vlan in self.vlans:
            elem.append(vlan.to_xml_element())
        return elem



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock
from xml.etree import ElementTree

from oslotest import base

from networking_baremetal import constants
from networking_baremetal.openconfig.vlan import vlan


class TestVlan(base.BaseTestCase):

    @mock.patch.object(vlan, 'Vlan', autospec=True)
    def test_vlans(self, mock_vlan):
        mock_vlan.return_value.to_xml_element.side_effect = [
            ElementTree.Element('fake-vlan-10'),
            ElementTree.Element('fake-vlan-20')
        ]
        oc_vlans = vlan.Vlans()
        oc_vlan = oc_vlans.add(10)
        mock_vlan.assert_called_with(10)
        self.assertEqual([oc_vlan], oc_vlans.vlans)
        remove_vlan = oc_vlans.remove(20)
        self.assertEqual([oc_vlan, remove_vlan], oc_vlans.vlans)
        element = oc_vlans.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = (f'<vlans xmlns="{oc_vlans.NAMESPACE}">'
                    '<fake-vlan-10 />'
                    '<fake-vlan-20 />'
                    '</vlans>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(vlan, 'VlanConfig', autospec=True)
    def test_vlan(self, mock_vlan_conf):
        mock_vlan_conf.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-vlan-conf'))
        oc_vlan = vlan.Vlan(10)
        self.assertEqual(constants.NetconfEditConfigOperation.MERGE.value,
                         oc_vlan.operation)
        self.assertEqual(10, oc_vlan.vlan_id)
        self.assertRaises(TypeError,
                          vlan.Vlan, 'not-int')
        self.assertRaises(ValueError, vlan.Vlan, 20,
                          **dict(operation='invalid'))
        element = oc_vlan.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<vlan>'
                    '<vlan-id operation="merge">10</vlan-id>'
                    '<fake-vlan-conf />'
                    '</vlan>')
        self.assertEqual(expected, xml_str)
        oc_vlan.operation = 'remove'
        element = oc_vlan.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<vlan>'
                    '<vlan-id operation="remove">10</vlan-id>'
                    '<fake-vlan-conf />'
                    '</vlan>')
        self.assertEqual(expected, xml_str)

    def test_vlan_config(self):
        vlan_conf = vlan.VlanConfig()
        self.assertEqual(constants.NetconfEditConfigOperation.MERGE.value,
                         vlan_conf.operation)
        self.assertRaises(ValueError, vlan.VlanConfig,
                          **dict(operation='invalid'))
        self.assertRaises(TypeError, vlan.VlanConfig,
                          **dict(vlan_id='not-int'))
        self.assertRaises(TypeError, vlan.VlanConfig,
                          **dict(name=20))  # Not str
        self.assertRaises(ValueError, vlan.VlanConfig,
                          **dict(status='invalid'))
        vlan_conf.vlan_id = 10
        vlan_conf.name = 'Vlan10'
        vlan_conf.status = 'ACTIVE'
        element = vlan_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config operation="merge">'
                    '<vlan-id>10</vlan-id>'
                    '<name>Vlan10</name>'
                    '<status>ACTIVE</status>'
                    '</config>')
        self.assertEqual(expected, xml_str)
        vlan_conf.operation = 'delete'
        vlan_conf.status = 'SUSPENDED'
        element = vlan_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config operation="delete">'
                    '<vlan-id>10</vlan-id>'
                    '<name>Vlan10</name>'
                    '<status>SUSPENDED</status>'
                    '</config>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(vlan, 'VlanSwitchedConfig', autospec=True)
    def test_switched_vlan(self, mock_switched_vlan_conf):
        mock_switched_vlan_conf.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-switched-vlan-config'))
        switched_vlan = vlan.VlanSwitchedVlan()
        element = switched_vlan.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = (f'<switched-vlan xmlns="{switched_vlan.NAMESPACE}">'
                    '<fake-switched-vlan-config />'
                    '</switched-vlan>')
        self.assertEqual(expected, xml_str)

    def test_switched_vlan_config(self):
        swithced_vlan_conf = vlan.VlanSwitchedConfig()
        self.assertEqual(constants.NetconfEditConfigOperation.MERGE.value,
                         swithced_vlan_conf.operation)
        self.assertRaises(ValueError, vlan.VlanSwitchedConfig,
                          **dict(operation='invalid'))
        self.assertRaises(ValueError, vlan.VlanSwitchedConfig,
                          **dict(interface_mode='invalid'))
        self.assertRaises(TypeError, vlan.VlanSwitchedConfig,
                          **dict(native_vlan='not-int'))
        self.assertRaises(TypeError, vlan.VlanSwitchedConfig,
                          **dict(access_vlan='not-int'))
        swithced_vlan_conf.interface_mode = 'ACCESS'
        swithced_vlan_conf.access_vlan = 20
        element = swithced_vlan_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config operation="merge">'
                    '<interface-mode>ACCESS</interface-mode>'
                    '<access-vlan>20</access-vlan>'
                    '</config>')
        self.assertEqual(expected, xml_str)
        del swithced_vlan_conf.access_vlan
        swithced_vlan_conf.interface_mode = 'TRUNK'
        swithced_vlan_conf.native_vlan = 30
        swithced_vlan_conf.trunk_vlans.add('10..50')
        swithced_vlan_conf.trunk_vlans.add('99')
        swithced_vlan_conf.trunk_vlans.add(88)
        swithced_vlan_conf.trunk_vlans.add('200..300')
        element = swithced_vlan_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config operation="merge">'
                    '<interface-mode>TRUNK</interface-mode>'
                    '<native-vlan>30</native-vlan>'
                    '<trunk-vlans>10..50</trunk-vlans>'
                    '<trunk-vlans>99</trunk-vlans>'
                    '<trunk-vlans>88</trunk-vlans>'
                    '<trunk-vlans>200..300</trunk-vlans>'
                    '</config>')
        self.assertEqual(expected, xml_str)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import random
import re
from urllib.parse import parse_qs as urlparse_qs
from urllib.parse import urlparse
import uuid
from xml.etree import ElementTree

from ncclient import manager
from ncclient.operations.rpc import RPCError
from ncclient.transport.errors import AuthenticationError
from ncclient.transport.errors import SessionCloseError
from ncclient.transport.errors import SSHError
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net
from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exec
from neutron_lib.plugins.ml2 import api
from oslo_config import cfg
from oslo_log import log as logging
import tenacity

from networking_baremetal import common
from networking_baremetal import config
from networking_baremetal import constants
from networking_baremetal.constants import NetconfEditConfigOperation as nc_op
from networking_baremetal.drivers import base
from networking_baremetal import exceptions
from networking_baremetal.openconfig.interfaces import interfaces
from networking_baremetal.openconfig.lacp import lacp
from networking_baremetal.openconfig.network_instance import network_instance
from networking_baremetal.openconfig.vlan import vlan

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

LOCK_DENIED_TAG = 'lock-denied'  # [RFC 4741]
CANDIDATE = 'candidate'
RUNNING = 'running'
DEFERRED = 'deferred'

# Options for the device, maps to the local_link_information in the
# port binding profile.
_DEVICE_OPTS = [
    cfg.StrOpt('network_instance',
               default='default',
               advanced=True,
               help=('The L2, L3, or L2+L3 forwarding instance to use when '
                     'defining VLANs on the device.')),
    cfg.DictOpt('port_id_re_sub',
                default={},
                sample_default={'pattern': 'Ethernet', 'repl': 'eth'},
                help=('Regular expression pattern and replacement string. '
                      'Some devices do not use the port description from '
                      'LLDP in Netconf configuration. If the regular '
                      'expression pattern and replacement string is set the '
                      'port_id will be modified before passing configuration '
                      'to the device.')),
    cfg.ListOpt('disabled_properties',
                item_type=cfg.types.String(
                    choices=['port_mtu']),
                default=[],
                help=('A list of properties that should not be used, '
                      'currently only "port_mtu" is valid')),
    cfg.BoolOpt('manage_lacp_aggregates',
                default=True,
                help=('When set to true the driver will manage LACP '
                      'aggregates if link_group_information is defined in '
                      'the binding:profile. When this is false the driver '
                      'expect the link aggregation to be pre-configured on '
                      'the device, and only perform vlan plugging.')),
    cfg.StrOpt('link_aggregate_prefix',
               default='Port-Channel',
               help=('The device specific prefix used for link-aggregation '
                     'ports. Common values: "po", "port-channel" or '
                     '"Port-Channel".')),
    cfg.StrOpt('link_aggregate_range',
               default='1000..2000',
               help=('Range of link aggregation interface IDs that the driver '
                     'can use when managing link aggregates.')),
]

# Configuration option for Netconf client connection
_NCCLIENT_OPTS = [
    cfg.StrOpt('host',
               help=('The hostname or IP address to use for connecting to the '
                     'netconf device.'),
               sample_default='device.example.com'),
    cfg.StrOpt('username',
               help='The username to use for SSH authentication.',
               sample_default='netconf'),
    cfg.IntOpt('port', default=830,
               help=('The port to use for connection to the netconf '
                     'device.')),
    cfg.StrOpt('password',
               help=('The password used if using password authentication, or '
                     'the passphrase to use for unlocking keys that require '
                     'it. (To disable attempting key authentication '
                     'altogether, set options *allow_agent* and '
                     '*look_for_keys* to `False`.'),
               sample_default='secret'),
    cfg.StrOpt('key_filename',
               help='Private key filename',
               default='~/.ssh/id_rsa'),
    cfg.BoolOpt('hostkey_verify',
                default=True,
                help=('Enables hostkey verification from '
                      '~/.ssh/known_hosts')),
    cfg.DictOpt('device_params',
                default={'name': 'default'},
                help=('ncclient device handler parameters, see ncclient '
                      'documentation for supported device handlers.')),
    cfg.BoolOpt('allow_agent',
                default=True,
                help='Enables querying SSH agent (if found) for keys.'),
    cfg.BoolOpt('look_for_keys',
                default=True,
                help=('Enables looking in the usual locations for ssh keys '
                      '(e.g. :file:`~/.ssh/id_*`)')),
]


def list_driver_opts():
    return [('networking_baremetal', config._opts),
            ('netconf-openconfig-example',
             config._device_opts + _DEVICE_OPTS + _NCCLIENT_OPTS)]


class NetconfLockDenied(n_exec.NeutronException):
    message = ('Access to the requested lock is denied because the'
               'lock is currently held by another entity.')


class NetconfOpenConfigClient(base.BaseDeviceClient):

    def __init__(self, device):
        super().__init__(device)
        self.device = device
        self.capabilities = set()

        # Reduce the log level for ncclient, it is very chatty by default
        netconf_logger = logging.getLogger('ncclient')
        netconf_logger.setLevel(logging.WARNING)

    @staticmethod
    def _get_lock_session_id(err_info):
        """Parse lock-denied error [RFC6241]

        error-tag:      lock-denied
        error-type:     protocol
        error-severity: error
        error-info:     <session-id> : session ID of session holding the
                        requested lock, or zero to indicate a non-NETCONF
                        entity holds the lock
        Description:    Access to the requested lock is denied because the
                        lock is currently held by another entity.
        """
        root = ElementTree.fromstring(err_info)
        session_id = root.find(
            "./{urn:ietf:params:xml:ns:netconf:base:1.0}session-id").text

        return session_id

    @staticmethod
    def process_capabilities(server_capabilities):
        capabilities = set()
        for capability in server_capabilities:
            for k, v in constants.IANA_NETCONF_CAPABILITIES.items():
                if v in capability:
                    capabilities.add(k)
            if capability.startswith('http://openconfig.net/yang'):
                openconfig_module = urlparse_qs(
                    urlparse(capability).query).get('module').pop()
                capabilities.add(openconfig_module)

        return capabilities

    def get_capabilities(self):
        # https://github.com/ncclient/ncclient/issues/525
        _ignore_close_issue_525 = False
        args = self.get_client_args()
        try:
            with manager.connect(**args) as nc_client:
                server_capabilities = nc_client.server_capabilities
                _ignore_close_issue_525 = True
        except SessionCloseError as e:
            if not _ignore_close_issue_525:
                raise e
        except (SSHError, AuthenticationError) as e:
            raise exceptions.DeviceConnectionError(device=self.device, err=e)

        return self.process_capabilities(server_capabilities)

    def get_client_args(self):
        """Get client connection arguments from configuration

        :param device: Device identifier
        """
        args = dict(
            host=CONF[self.device].host,
            port=CONF[self.device].port,
            username=CONF[self.device].username,
            hostkey_verify=CONF[self.device].hostkey_verify,
            device_params=CONF[self.device].device_params,
            keepalive=True,
            allow_agent=CONF[self.device].allow_agent,
            look_for_keys=CONF[self.device].look_for_keys,
        )
        if CONF[self.device].key_filename:
            args['key_filename'] = CONF[self.device].key_filename
        if CONF[self.device].password:
            args['password'] = CONF[self.device].password

        return args

    def get(self, **kwargs):
        """Get current configuration/staate from device"""
        # https://github.com/ncclient/ncclient/issues/525
        _ignore_close_issue_525 = False

        query = kwargs.get('query')
        q_filter = ElementTree.tostring(query.to_xml_element()).decode('utf-8')
        try:
            with manager.connect(**self.get_client_args()) as client:
                reply = client.get(filter=('subtree', q_filter))
                _ignore_close_issue_525 = True
        except SessionCloseError as e:
            # https://github.com/ncclient/ncclient/issues/525
            if not _ignore_close_issue_525:
                raise e
        except RPCError as e:
            LOG.error('Netconf XML: %s', q_filter)
            raise e

        return reply.data_xml

    @tenacity.retry(
        reraise=True,
        retry=tenacity.retry_if_exception_type(NetconfLockDenied),
        wait=tenacity.wait_random_exponential(multiplier=1, min=2, max=10),
        stop=tenacity.stop_after_attempt(5))
    def get_lock_and_configure(self, client, source, config,
                               deferred_allocations):
        try:
            with client.locked(source):
                # Aggregate ID deferred until we have config lock
                # Get free aggregate ID by querying the device and update conf
                if deferred_allocations:
                    aggregate_id = self.get_free_aggregate_id(client)
                    self.allocate_deferred(aggregate_id, config)
                xml_config = common.config_to_xml(config)
                LOG.info(
                    'Sending configuration to Netconf device %(dev)s: '
                    '%(conf)s',
                    {'dev': self.device, 'conf': xml_config})
                if source == CANDIDATE:
                    # Revert the candidate configuration to the current
                    # running configuration. Any uncommitted changes are
                    # discarded.
                    client.discard_changes()
                    # Edit the candidate configuration
                    client.edit_config(target=source, config=xml_config)
                    # Validate the candidate configuration
                    if (':validate' in self.capabilities
                            or ':validate:1.1' in self.capabilities):
                        client.validate(source='candidate')
                    # Commit the candidate config, 30 seconds timeout
                    if (':confirmed-commit' in self.capabilities
                            or ':confirmed-commit:1.1' in self.capabilities):
                        client.commit(confirmed=True, timeout=str(30))
                        # Confirm the commit, if this commit does not
                        # succeed the device will revert the config after
                        # 30 seconds.
                    client.commit()
                elif source == RUNNING:
                    client.edit_config(target=source, config=xml_config)
                # TODO(hjensas): persist config.
        except RPCError as err:
            if err.tag == LOCK_DENIED_TAG:
                # If the candidate config is modified, some vendors do not
                # permit a new session to take a lock. This is per the RFC,
                # in this case a lock-denied error where session-id == 0 is
                # returned, because no session is actually holding the
                # lock we can discard changes which will release the lock.
                if (source == CANDIDATE
                        and self._get_lock_session_id(err.info) == '0'):
                    client.discard_changes()
                raise NetconfLockDenied()
            else:
                LOG.error('Netconf XML: %s', common.config_to_xml(config))
                raise err

    def edit_config(self, config, deferred_allocations=False):
        """Edit configuration on the device

        :param config: Configuration, or list of configurations
        :param deferred_allocations: Used for link aggregates, the aggregate
          id cannot be allocated before device config is locked. When this
          is true an available aggregate id is identified by querying the
          device, and the configuration objects are updated accordingly before
          configuration is sent to the device.
        """

        # https://github.com/ncclient/ncclient/issues/525
        _ignore_close_issue_525 = False

        if not isinstance(config, list):
            config = [config]

        try:
            with manager.connect(**self.get_client_args()) as client:
                self.capabilities = self.process_capabilities(
                    client.server_capabilities)
                if ':candidate' in self.capabilities:
                    self.get_lock_and_configure(
                        client, CANDIDATE, config, deferred_allocations)
                    _ignore_close_issue_525 = True
                elif ':writable-running' in self.capabilities:
                    self.get_lock_and_configure(
                        client, RUNNING, config, deferred_allocations)
                    _ignore_close_issue_525 = True
        except SessionCloseError as e:
            if not _ignore_close_issue_525:
                raise e

    def get_aggregation_ids(self):
        """Get aggregation IDs and aggregation prefix from config"""
        prefix = CONF[self.device].link_aggregate_prefix
        aggregate_id_range = CONF[self.device].link_aggregate_range.split('..')
        aggregate_ids = {f'{prefix}{x}'
                         for x in range(int(aggregate_id_range[0]),
                                        int(aggregate_id_range[1]) + 1)}
        return aggregate_ids

    @staticmethod
    def allocate_deferred(aggregate_id, config):
        """Set aggregation id where it was deffered

        :param aggregate_id: Aggregation ID for the link aggregate,
            for example 'po123'
        :param config: Configuration objects to update
        """
        for conf in config:
            if isinstance(conf, interfaces.Interfaces):
                for iface in conf:
                    if isinstance(iface, interfaces.InterfaceAggregate):
                        if iface.name == DEFERRED:
                            iface.name = aggregate_id
                        if iface.config.name == DEFERRED:
                            iface.config.name = aggregate_id
                    elif isinstance(iface, interfaces.InterfaceEthernet):
                        if iface.ethernet.config.aggregate_id == DEFERRED:
                            iface.ethernet.config.aggregate_id = aggregate_id
            if isinstance(conf, lacp.LACP):
                for lacp_iface in conf.interfaces.interfaces:
                    if lacp_iface.name == DEFERRED:
                        lacp_iface.name = aggregate_id

    def get_free_aggregate_id(self, client_locked):
        """Get free aggregate id by querying device config

        :param client_locked: Netconf client with active
            configuration lock
        """
        aggregate_prefix = CONF[self.device].link_aggregate_prefix
        aggregate_ids = self.get_aggregation_ids()
        # Create a interfaces query
        oc_ifaces = interfaces.Interfaces()
        # Use empty string for the name, so the 'get' return all interfaces
        oc_iface = oc_ifaces.add('', interface_type=constants.IFACE_TYPE_BASE)
        # Don't need the config group
        del oc_iface.config
        # Get interfaces from device
        element = oc_ifaces.to_xml_element()
        device_interfaces = client_locked.get(filter=(
            'subtree', ElementTree.tostring(element).decode("utf-8")))
        # Find all interface names and filter on aggregate_prefix
        root = ElementTree.fromstring(device_interfaces.data_xml)
        used_aggregate_ids = {
            x.text for x in root.findall(f'.//{{{oc_ifaces.NAMESPACE}}}name')
            if x.text.startswith(aggregate_prefix)}
        # Get the difference, and make a random choice
        available_aggregate_ids = aggregate_ids.difference(used_aggregate_ids)

        return random.choice(list(available_aggregate_ids))


class NetconfOpenConfigDriver(base.BaseDeviceDriver):

    SUPPORTED_BOND_MODES = set().union(constants.NON_SWITCH_BOND_MODES,
                                       constants.LACP_BOND_MODES,
                                       constants.PRE_CONF_ONLY_BOND_MODES)

    def __init__(self, device):
        super().__init__(device)
        self.client = NetconfOpenConfigClient(device)
        self.device = device

    def validate(self):
        try:
            LOG.info('Device %(device)s was loaded. Device capabilities: '
                     '%(caps)s', {'device': self.device,
                                  'caps': self.client.get_capabilities()})
        except exceptions.DeviceConnectionError as e:
            raise exceptions.DriverValidationError(device=self.device, err=e)

    def load_config(self):
        """Register driver specific configuration"""

        CONF.register_opts(_DEVICE_OPTS, group=self.device)
        CONF.register_opts(_NCCLIENT_OPTS, group=self.device)

    def create_network(self, context):
        """Create network on device

        :param context: NetworkContext instance describing the new
            network.
        """
        network = context.current
        segmentation_id = network[provider_net.SEGMENTATION_ID]

        net_instances = network_instance.NetworkInstances()
        net_instance = net_instances.add(CONF[self.device].network_instance)
        _vlan = net_instance.vlans.add(segmentation_id)
        # Devices has limitations for vlan names, use the hex variant of the
        # network UUID which is shorter.
        _vlan.config.name = self._uuid_as_hex(network[api.ID])
        _vlan.config.status = constants.VLAN_ACTIVE
        self.client.edit_config(net_instances)

    def update_network(self, context):
        """Update network on device

        :param context: NetworkContext instance describing the new
            network.
        """
        network = context.current
        network_orig = context.original
        segmentation_id = network[provider_net.SEGMENTATION_ID]
        segmentation_id_orig = network_orig[provider_net.SEGMENTATION_ID]
        admin_state = network['admin_state_up']
        admin_state_orig = network_orig['admin_state_up']

        add_net_instances = network_instance.NetworkInstances()
        add_net_instance = add_net_instances.add(
            CONF[self.device].network_instance)
        del_net_instances = None
        need_update = False

        if segmentation_id:
            _vlan = add_net_instance.vlans.add(segmentation_id)
            # Devices has limitations for vlan names, use the hex variant of
            # the network UUID which is shorter.
            _vlan.config.name = self._uuid_as_hex(network[api.ID])
            if network['admin_state_up']:
                _vlan.config.status = constants.VLAN_ACTIVE
            else:
                _vlan.config.status = constants.VLAN_SUSPENDED
            if admin_state != admin_state_orig:
                need_update = True
        if segmentation_id_orig and segmentation_id != segmentation_id_orig:
            need_update = True
            del_net_instances = network_instance.NetworkInstances()
            del_net_instance = del_net_instances.add(
                CONF[self.device].network_instance)
            vlan_orig = del_net_instance.vlans.remove(segmentation_id_orig)
            # Not all devices support removing a VLAN, in that case lets
            # make sure the VLAN is suspended and set a name to indicate the
            # network was deleted.
            vlan_orig.config.name = f'neutron-DELETED-{segmentation_id_orig}'
            vlan_orig.config.status = constants.VLAN_SUSPENDED

        if not need_update:
            return

        # If the segmentation ID changed, delete the old VLAN first to avoid
        # vlan name conflict.
        if del_net_instances is not None:
            self.client.edit_config(del_net_instances)

        self.client.edit_config(add_net_instances)

    def delete_network(self, context):
        """Delete network on device

        :param context: NetworkContext instance describing the new
            network.
        """
        network = context.current
        segmentation_id = network[provider_net.SEGMENTATION_ID]

        net_instances = network_instance.NetworkInstances()
        net_instance = net_instances.add(CONF[self.device].network_instance)
        _vlan = net_instance.vlans.remove(segmentation_id)
        # Not all devices support removing a VLAN, in that case lets
        # make sure the VLAN is suspended and set a name to indicate the
        # network was deleted.
        _vlan.config.name = f'neutron-DELETED-{segmentation_id}'
        _vlan.config.status = constants.VLAN_SUSPENDED
        self.client.edit_config(net_instances)

    def create_port(self, context, segment, links):
        """Create/Configure port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param segment: segment dictionary describing segment to bind
        :param links: Local link information filtered for the device.
        """
        port = context.current
        binding_profile = port[portbindings.PROFILE]
        local_group_information = binding_profile.get(
            constants.LOCAL_GROUP_INFO, {})
        bond_mode = local_group_information.get('bond_mode')

        if segment[api.NETWORK_TYPE] != n_const.TYPE_VLAN:
            switched_vlan = None
        else:
            switched_vlan = vlan.VlanSwitchedVlan()
            switched_vlan.config.operation = nc_op.REPLACE
            switched_vlan.config.interface_mode = constants.VLAN_MODE_ACCESS
            switched_vlan.config.access_vlan = segment[api.SEGMENTATION_ID]

        if not bond_mode or bond_mode in constants.NON_SWITCH_BOND_MODES:
            self.create_non_bond(context, switched_vlan, links)
        elif bond_mode in constants.LACP_BOND_MODES:
            if CONF[self.device].manage_lacp_aggregates:
                self.create_lacp_aggregate(context, switched_vlan, links)
            else:
                self.create_pre_conf_aggregate(context, switched_vlan, links)
        elif bond_mode in constants.PRE_CONF_ONLY_BOND_MODES:
            self.create_pre_conf_aggregate(context, switched_vlan, links)

    def create_non_bond(self, context, switched_vlan, links):
        """Create/Configure ports on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param switched_vlan: switched_vlan OpenConfig object
        :param links: Local link information filtered for the device.
        """
        port = context.current
        network = context.network.current

        ifaces = interfaces.Interfaces()
        for link in links:
            link_port_id = link.get(constants.PORT_ID)
            link_port_id = self._port_id_resub(link_port_id)

            iface = ifaces.add(link_port_id)
            iface.config.enabled = port['admin_state_up']
            if 'port_mtu' not in CONF[self.device].disabled_properties:
                iface.config.mtu = network[api.MTU]
            iface.config.description = f'neutron-{port[api.ID]}'
            if switched_vlan is not None:
                iface.ethernet.switched_vlan = switched_vlan
            else:
                del iface.ethernet

        self.client.edit_config(ifaces)

    def create_lacp_aggregate(self, context, switched_vlan, links):
        """Create/Configure LACP aggregate on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param switched_vlan: switched_vlan OpenConfig object
        :param links: Local link information filtered for the device.
        """
        port = context.current
        network = context.network.current
        binding_profile = port[portbindings.PROFILE]
        local_group_information = binding_profile.get(
            constants.LOCAL_GROUP_INFO, {})
        dev_type = CONF[self.device].device_params.get('name')
        bond_properties = local_group_information.get('bond_properties', {})
        lacp_interval = bond_properties.get(constants.LACP_INTERVAL)
        min_links = bond_properties.get(constants.LACP_MIN_LINKS)
        ifaces = interfaces.Interfaces()
        _lacp = lacp.LACP()
        lacp_iface = _lacp.interfaces.add(DEFERRED)
        lacp_iface.operation = nc_op.REPLACE
        lacp_iface.config.interval = (constants.LACP_PERIOD_FAST
                                      if lacp_interval in {'fast', 1, '1'}
                                      else constants.LACP_PERIOD_SLOW)
        # NX-API only allows configuring LACP interval rate on a port-channel
        # member which is not in shutdown state. Support would require a two
        # commit approach.
        if dev_type in {'nexus'}:
            LOG.warning('IGNORING LACP interval (bond_lacp_rate). The driver '
                        'does not support LACP interval for this device type. '
                        'Device: %(device)s, Port: %(port)s',
                        {'device': self.device, 'port': port[api.ID]})
            del lacp_iface.config.interval

        for link in links:
            link_port_id = link.get(constants.PORT_ID)
            link_port_id = self._port_id_resub(link_port_id)
            iface = ifaces.add(
                link_port_id, interface_type=constants.IFACE_TYPE_ETHERNET)
            iface.config.operation = nc_op.MERGE
            iface.config.enabled = port['admin_state_up']
            if 'port_mtu' not in CONF[self.device].disabled_properties:
                iface.config.mtu = network[api.MTU]
            iface.config.description = f'neutron-{port[api.ID]}'
            iface.ethernet.config.aggregate_id = DEFERRED

        iface = ifaces.add(DEFERRED,
                           interface_type=constants.IFACE_TYPE_AGGREGATE)
        iface.config.operation = nc_op.MERGE
        iface.config.name = DEFERRED
        iface.config.enabled = port['admin_state_up']
        iface.config.description = f'neutron-{port[api.ID]}'
        iface.aggregation.config.lag_type = constants.LAG_TYPE_LACP
        if min_links:
            iface.aggregation.config.min_links = int(min_links)
        if switched_vlan is not None:
            iface.aggregation.switched_vlan = switched_vlan
        else:
            del iface.aggregation.switched_vlan

        self.client.edit_config([ifaces, _lacp], deferred_allocations=True)

    def create_pre_conf_aggregate(self, context, switched_vlan, links):
        """Create/Configure pre-configured aggregate on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param switched_vlan: switched_vlan OpenConfig object
        :param links: Local link information filtered for the device.
        """
        port = context.current
        aggregate_ids = self.get_aggregate_ids(links)
        if not aggregate_ids:
            raise exceptions.PreConfiguredAggrergateNotFound(
                links=links, device=self.device)
        ifaces = interfaces.Interfaces()
        for aggregate_id in aggregate_ids:
            iface = ifaces.add(aggregate_id,
                               interface_type=constants.IFACE_TYPE_AGGREGATE)
            iface.operation = nc_op.MERGE
            iface.config.enabled = port['admin_state_up']
            if switched_vlan is not None:
                iface.aggregation.switched_vlan = switched_vlan
            else:
                del iface.aggregation.switched_vlan

        self.client.edit_config(ifaces)

    def update_port(self, context, links):
        """Update port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        """
        if (not self.admin_state_changed(context)
                and not self.network_mtu_changed(context)):
            return

        port = context.current
        binding_profile = port[portbindings.PROFILE]
        local_group_information = binding_profile.get(
            constants.LOCAL_GROUP_INFO, {})
        bond_mode = local_group_information.get('bond_mode')

        if not bond_mode or bond_mode in constants.NON_SWITCH_BOND_MODES:
            self.update_non_bond(context, links)
        elif bond_mode in constants.LACP_BOND_MODES:
            if CONF[self.device].manage_lacp_aggregates:
                self.update_lacp_aggregate(context, links)
            else:
                self.update_pre_conf_aggregate(context, links)
        elif bond_mode in constants.PRE_CONF_ONLY_BOND_MODES:
            self.update_pre_conf_aggregate(context, links)

    def update_non_bond(self, context, links):
        """Update port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        """
        network = context.network.current
        ifaces = interfaces.Interfaces()
        port = context.current

        for link in links:
            link_port_id = link.get(constants.PORT_ID)
            link_port_id = self._port_id_resub(link_port_id)

            iface = ifaces.add(link_port_id)
            iface.config.enabled = port['admin_state_up']
            if 'port_mtu' not in CONF[self.device].disabled_properties:
                iface.config.mtu = network[api.MTU]

            del iface.ethernet

        self.client.edit_config(ifaces)

    def update_lacp_aggregate(self, context, links):
        """Update LACP aggregate on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        """
        port = context.current
        network = context.network.current
        aggregate_ids = self.get_aggregate_ids(links)
        ifaces = interfaces.Interfaces()
        for link in links:
            link_port_id = link.get(constants.PORT_ID)
            link_port_id = self._port_id_resub(link_port_id)

            iface = ifaces.add(link_port_id,
                               interface_type=constants.IFACE_TYPE_ETHERNET)
            iface.config.enabled = port['admin_state_up']
            if 'port_mtu' not in CONF[self.device].disabled_properties:
                iface.config.mtu = network[api.MTU]

            del iface.ethernet

        for aggregate_id in aggregate_ids:
            iface = ifaces.add(aggregate_id,
                               interface_type=constants.IFACE_TYPE_AGGREGATE)
            iface.operation = nc_op.MERGE
            iface.config.enabled = port['admin_state_up']
            del iface.aggregation

        self.client.edit_config(ifaces)

    def update_pre_conf_aggregate(self, context, links):
        """Update pre-configured aggregate on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        """
        port = context.current
        aggregate_ids = self.get_aggregate_ids(links)
        if not aggregate_ids:
            raise exceptions.PreConfiguredAggrergateNotFound(
                links=links, device=self.device)
        ifaces = interfaces.Interfaces()
        for aggregate_id in aggregate_ids:
            iface = ifaces.add(aggregate_id,
                               interface_type=constants.IFACE_TYPE_AGGREGATE)
            iface.operation = nc_op.MERGE
            iface.config.enabled = port['admin_state_up']

        self.client.edit_config(ifaces)

    def delete_port(self, context, links, current=True):
        """Delete/Un-configure port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        :param current: Boolean, when true use context.current, when
            false use context.original
        """
        port = context.current if current else context.original
        binding_profile = port[portbindings.PROFILE]
        local_group_information = binding_profile.get(
            constants.LOCAL_GROUP_INFO, {})
        bond_mode = local_group_information.get('bond_mode')

        if not bond_mode or bond_mode in constants.NON_SWITCH_BOND_MODES:
            self.delete_non_bond(context, links)
        elif bond_mode in constants.LACP_BOND_MODES:
            if CONF[self.device].manage_lacp_aggregates:
                self.delete_lacp_aggregate(context, links)
            else:
                self.delete_pre_conf_aggregate(links)
        elif bond_mode in constants.PRE_CONF_ONLY_BOND_MODES:
            self.delete_pre_conf_aggregate(links)

    def delete_non_bond(self, context, links):
        """Delete/Un-configure port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        """
        network = context.network.current
        ifaces = interfaces.Interfaces()
        for link in links:
            link_port_id = link.get(constants.PORT_ID)
            link_port_id = self._port_id_resub(link_port_id)

            iface = ifaces.add(link_port_id)
            iface.config.operation = nc_op.REMOVE
            # Not possible mark entire config for removal due to name leaf-ref
            # Set dummy values for properties to remove
            iface.config.description = ''
            iface.config.enabled = False
            if 'port_mtu' not in CONF[self.device].disabled_properties:
                iface.config.mtu = 0
            if network[provider_net.NETWORK_TYPE] == n_const.TYPE_VLAN:
                iface.ethernet.switched_vlan.config.operation = nc_op.REMOVE
            else:
                del iface.ethernet

        self.client.edit_config(ifaces)

    def delete_lacp_aggregate(self, context, links):
        """Delete/Un-configure LACP aggregate on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        """
        network = context.network.current
        aggregate_ids = self.get_aggregate_ids(links)
        ifaces = interfaces.Interfaces()
        for link in links:
            link_port_id = link.get(constants.PORT_ID)
            link_port_id = self._port_id_resub(link_port_id)
            # Set up interface links for config remove
            iface = ifaces.add(link_port_id,
                               interface_type=constants.IFACE_TYPE_ETHERNET)
            iface.config.operation = nc_op.REMOVE
            iface.config.description = ''
            iface.config.enabled = False
            if 'port_mtu' not in CONF[self.device].disabled_properties:
                iface.config.mtu = 0
            iface.ethernet.config.operation = nc_op.REMOVE
            if network[provider_net.NETWORK_TYPE] == n_const.TYPE_VLAN:
                iface.ethernet.switched_vlan.config.operation = nc_op.REMOVE
            else:
                del iface.ethernet.switched_vlan

        # Set up lacp and aggregate interface for removal
        _lacp = lacp.LACP()
        for aggregate_id in aggregate_ids:
            # Remove LACP interface
            lacp_iface = _lacp.interfaces.add(aggregate_id)
            lacp_iface.operation = nc_op.REMOVE
            del lacp_iface.config
            # Remove Aggregate interface
            iface = ifaces.add(aggregate_id,
                               interface_type=constants.IFACE_TYPE_AGGREGATE)
            iface.operation = nc_op.REMOVE
            del iface.config
            del iface.aggregation

        self.client.edit_config([_lacp, ifaces])

    def delete_pre_conf_aggregate(self, links):
        """Delete/Un-configure pre-configured aggregate on device

        :param links: Local link information filtered for the device.
        """
        aggregate_ids = self.get_aggregate_ids(links)
        if not aggregate_ids:
            raise exceptions.PreConfiguredAggrergateNotFound(
                links=links, device=self.device)
        ifaces = interfaces.Interfaces()
        for aggregate_id in aggregate_ids:
            iface = ifaces.add(aggregate_id,
                               interface_type=constants.IFACE_TYPE_AGGREGATE)
            iface.config.enabled = False
            iface.aggregation.switched_vlan.config.operation = nc_op.REMOVE

        self.client.edit_config(ifaces)

    @staticmethod
    def _uuid_as_hex(_uuid):
        return uuid.UUID(_uuid).hex

    def _port_id_resub(self, link_port_id):
        """Replace pattern

        Regular expression pattern and replacement string.
        Some devices don not use the port  description from
        LLDP in Netconf configuration. If the regular expression
        pattern and replacement string is set the  port_id will
        be modified before passing configuration to the device.

        Replacing the leftmost non-overlapping occurrences of pattern
        in string by the replacement repl.
        """
        if CONF[self.device].port_id_re_sub:
            pattern = CONF[self.device].port_id_re_sub.get('pattern')
            repl = CONF[self.device].port_id_re_sub.get('repl')
            link_port_id = re.sub(pattern, repl, link_port_id)

        return link_port_id

    def get_aggregate_ids(self, links):
        query = interfaces.Interfaces()
        for link in links:
            link_port_id = link.get(constants.PORT_ID)
            link_port_id = self._port_id_resub(link_port_id)
            # Set up query
            q_iface = query.add(link_port_id,
                                interface_type=constants.IFACE_TYPE_ETHERNET)
            # Remove config and ethernet for broad filter.
            del q_iface.config
            del q_iface.ethernet

        # Get aggregate ids by querying the link interfaces
        xml_result = self.client.get(query=query)
        root = ElementTree.fromstring(xml_result)
        xpath_query_result = root.findall(
            './/{http://openconfig.net/yang/interfaces/aggregate}'
            'aggregate-id')
        aggregate_ids = {x.text for x in xpath_query_result}

        return aggregate_ids

    @staticmethod
    def admin_state_changed(context):
        port = context.current
        port_orig = context.original
        return (port and port_orig
                and port['admin_state_up'] != port_orig['admin_state_up'])

    @staticmethod
    def network_mtu_changed(context):
        network = context.network.current
        network_orig = context.network.original
        return (network and network_orig
                and network[api.MTU] != network_orig[api.MTU])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import enum

from networking_baremetal import constants


class LACPPeriod(enum.Enum):
    """Defines the time between sending LACP messages

    reference "IEEE 802.3ad"
    FAST: Send LACP packets every second
    SLOW: Send LACP packets every 30 seconds
    """
    FAST = constants.LACP_PERIOD_FAST
    SLOW = constants.LACP_PERIOD_SLOW


class LACPActivity(enum.Enum):
    """Describes the LACP membership type

    Active or passive, of the interface in the aggregate.
    reference "IEEE 802.1AX-2008"

    ACTIVE:  Interface is an active member, i.e., will detect and
             maintain aggregates
    PASSIVE: Interface is a passive member, i.e., it participates
             with an active partner
    """
    ACTIVE = constants.LACP_ACTIVITY_ACTIVE
    PASSIVE = constants.LACP_ACTIVITY_PASSIVE



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc


class DriverEntrypointLoadError(n_exc.NeutronException):
    message = 'Failed to load entrypoint %(entry_point)s: %(err)s'


class DriverValidationError(n_exc.NeutronException):
    message = 'Failed driver validation for device %(device)s: %(err)s'


class DeviceConnectionError(n_exc.NeutronException):
    message = 'Driver failed connecting to device %(device)s: %(err)s'


class PreConfiguredAggrergateNotFound(n_exc.NeutronException):
    message = ('Driver could not find the aggregate ID for the pre-configured '
               'link aggregate for links %(links)s on device %(device)s.')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import enum
import re

from networking_baremetal import constants


class VlanStatus(enum.Enum):
    """VLAN Admin state

    ACTIVE: VLAN is active
    SUSPENDED: VLAN is inactive / suspended
    """
    ACTIVE = constants.VLAN_ACTIVE
    SUSPENDED = constants.VLAN_SUSPENDED


class VlanInterfaceMode(enum.Enum):
    """VLAN interface mode (trunk or access)"""
    TRUNK = constants.VLAN_MODE_TRUNK
    ACCESS = constants.VLAN_MODE_ACCESS


class VlanId:
    """Type definition representing a single-tagged VLAN"""

    def __init__(self, vlan_id: int):
        if not isinstance(vlan_id, int):
            raise TypeError('vlan_id must be integer, got {}'
                            .format(type(vlan_id)))
        if vlan_id not in constants.VLAN_RANGE:
            raise ValueError('Invalid vlan id: {vlan_id} not in {range}'
                             .format(vlan_id=vlan_id,
                                     range=constants.VLAN_RANGE))
        self._vlan_id = vlan_id

    @property
    def vlan_id(self):
        return self._vlan_id


class VlanRange:
    """Type definition representing a range of single-tagged VLANs.

    A range is specified as x..y where x and y are
    valid VLAN IDs (1 <= vlan-id <= 4094). The range is
    assumed to be inclusive, such that any VLAN-ID matching
    x <= VLAN-ID <= y falls within the range."
    """
    # range specified as [lower]..[upper]
    pattern = re.compile(
        ('^(409[0-4]|40[0-8][0-9]|[1-3][0-9]{3}|'
         '[1-9][0-9]{1,2}|[1-9])\\.\\.(409[0-4]|'
         '40[0-8][0-9]|[1-3][0-9]{3}|[1-9][0-9]{1,2}|'
         '[1-9])$'))

    def __init__(self, vlan_range: str):
        if not isinstance(vlan_range, str):
            raise TypeError('vlan_range must be string, got {}'
                            .format(type(vlan_range)))
        if not self.pattern.match(vlan_range):
            raise ValueError('Invalid VLAN range {}'.format(vlan_range))
        lower, _, upper = vlan_range.partition('..')
        if not int(lower) <= int(upper):
            raise ValueError('Invalid VLAN range {}'.format(vlan_range))
        self._vlan_range = vlan_range

    @property
    def vlan_range(self):
        return self._vlan_range



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock
from xml.etree import ElementTree

from oslotest import base

from networking_baremetal import constants
from networking_baremetal.openconfig.interfaces import aggregate
from networking_baremetal.openconfig.interfaces import ethernet
from networking_baremetal.openconfig.interfaces import interfaces
from networking_baremetal.openconfig.vlan import vlan


class TestInterfaces(base.BaseTestCase):

    @mock.patch.object(ethernet, 'InterfacesEthernetConfig', autospec=True)
    @mock.patch.object(vlan, 'VlanSwitchedVlan', autospec=True)
    def test_interfaces_ethernet(self, mock_sw_vlan, mock_eth_conf):
        mock_sw_vlan.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-switched-vlan'))
        mock_eth_conf.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-ethernet-config'))
        if_eth = ethernet.InterfacesEthernet()
        mock_sw_vlan.assert_called_with()
        element = if_eth.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = (f'<ethernet xmlns="{if_eth.NAMESPACE}">'
                    '<fake-ethernet-config />'
                    '<fake-switched-vlan />'
                    '</ethernet>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(aggregate, 'InterfacesAggregationConfig', autospec=True)
    @mock.patch.object(vlan, 'VlanSwitchedVlan', autospec=True)
    def test_interfaces_aggregate(self, mock_sw_vlan, mock_agg_conf):
        mock_sw_vlan.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-switched-vlan'))
        mock_agg_conf.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-aggregate-config'))
        if_aggregate = aggregate.InterfacesAggregation()
        mock_sw_vlan.assert_called_with()
        element = if_aggregate.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = (f'<aggregation xmlns="{if_aggregate.NAMESPACE}">'
                    '<fake-aggregate-config />'
                    '<fake-switched-vlan />'
                    '</aggregation>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(interfaces, 'InterfaceAggregate', autospec=True)
    @mock.patch.object(interfaces, 'InterfaceEthernet', autospec=True)
    def test_interfaces_interfaces(self, mock_iface_eth, mock_iface_aggregate):
        mock_iface_eth.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-ethernet'))
        mock_iface_aggregate.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-aggregate'))
        ifaces = interfaces.Interfaces()
        iface = ifaces.add('eth0/1')
        iface2 = ifaces.add('po10', interface_type='aggregate')
        mock_iface_eth.assert_called_with('eth0/1')
        mock_iface_aggregate.assert_called_with('po10')
        self.assertEqual([iface, iface2], ifaces.interfaces)
        element = ifaces.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = (f'<interfaces xmlns="{ifaces.NAMESPACE}">'
                    '<fake-ethernet />'
                    '<fake-aggregate />'
                    '</interfaces>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(ethernet, 'InterfacesEthernet', autospec=True)
    @mock.patch.object(interfaces, 'InterfaceConfig', autospec=True)
    def test_interfaces_interface_ethernet(self, mock_if_conf, mock_if_eth):
        mock_if_conf.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake_config'))
        mock_if_eth.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake_ethernet'))
        interface = interfaces.InterfaceEthernet('eth0/1')
        mock_if_conf.assert_called_with()
        mock_if_eth.assert_called_with()
        self.assertEqual('eth0/1', interface.name)
        self.assertEqual(mock_if_conf(), interface.config)
        self.assertEqual(mock_if_eth(), interface.ethernet)
        element = interface.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<interface>'
                    '<name>eth0/1</name>'
                    '<fake_config />'
                    '<fake_ethernet />'
                    '</interface>')
        self.assertEqual(expected, xml_str)
        not_string = 10
        self.assertRaises(TypeError,
                          interfaces.InterfaceEthernet, not_string)

    @mock.patch.object(aggregate, 'InterfacesAggregation', autospec=True)
    @mock.patch.object(interfaces, 'InterfaceConfig', autospec=True)
    def test_interfaces_interface_aggregate(self, mock_if_conf, mock_if_aggr):
        mock_if_conf.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake_config'))
        mock_if_aggr.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake_aggregation'))
        interface = interfaces.InterfaceAggregate('po10')
        mock_if_conf.assert_called_with()
        mock_if_aggr.assert_called_with()
        self.assertEqual('po10', interface.name)
        self.assertEqual(mock_if_conf(), interface.config)
        self.assertEqual(mock_if_aggr(), interface.aggregation)
        element = interface.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<interface operation="merge">'
                    '<name>po10</name>'
                    '<fake_config />'
                    '<fake_aggregation />'
                    '</interface>')
        self.assertEqual(expected, xml_str)
        not_string = 10
        self.assertRaises(TypeError,
                          interfaces.InterfaceEthernet, not_string)

    def test_interfaces_interface_config(self):
        if_conf = interfaces.InterfaceConfig()
        self.assertEqual(constants.NetconfEditConfigOperation.MERGE.value,
                         if_conf.operation)
        self.assertRaises(ValueError, interfaces.InterfaceConfig,
                          **dict(operation='invalid'))
        self.assertRaises(TypeError, interfaces.InterfaceConfig,
                          **dict(enabled='not_bool'))
        self.assertRaises(TypeError, interfaces.InterfaceConfig,
                          **dict(description=10))  # Not string
        self.assertRaises(TypeError, interfaces.InterfaceConfig,
                          **dict(mtu='not_int'))
        if_conf.name = 'test1'
        if_conf.enabled = True
        if_conf.description = 'Description'
        if_conf.mtu = 9000
        element = if_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config>'
                    '<name operation="merge">test1</name>'
                    '<description operation="merge">Description</description>'
                    '<enabled operation="merge">true</enabled>'
                    '<mtu operation="merge">9000</mtu>'
                    '</config>')
        self.assertEqual(expected, xml_str)
        del if_conf.name
        if_conf.operation = 'remove'
        if_conf.description = ''
        if_conf.mtu = 0
        if_conf.enabled = False
        element = if_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config>'
                    '<description operation="remove" />'
                    '<enabled operation="remove">false</enabled>'
                    '<mtu operation="remove">0</mtu>'
                    '</config>')
        self.assertEqual(expected, xml_str)

    def test_interfaces_interface_ethernet_config(self):
        eth_conf = ethernet.InterfacesEthernetConfig()
        self.assertEqual(constants.NetconfEditConfigOperation.MERGE.value,
                         eth_conf.operation)
        self.assertRaises(ValueError,
                          ethernet.InterfacesEthernetConfig,
                          **dict(operation='invalid'))
        eth_conf.aggregate_id = 'po100'
        element = eth_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config operation="merge">'
                    '<aggregate-id '
                    'xmlns="http://openconfig.net/yang/interfaces/aggregate"'
                    '>po100</aggregate-id>'
                    '</config>')
        self.assertEqual(expected, xml_str)
        eth_conf = ethernet.InterfacesEthernetConfig()
        eth_conf.operation = 'remove'
        element = eth_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = '<config operation="remove" />'
        self.assertEqual(expected, xml_str)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from collections import abc
from xml.etree import ElementTree

from networking_baremetal import common
from networking_baremetal import constants
from networking_baremetal.openconfig.lacp import types


class LACP:
    """LACP Top level

    LACP configuration and state variable containers
    """

    NAMESPACE = 'http://openconfig.net/yang/lacp'
    TAG = 'lacp'

    def __init__(self):
        self._config = None
        self._interfaces = LACPInterfaces()

    @property
    def interfaces(self):
        return self._interfaces

    @interfaces.setter
    def interfaces(self, value):
        if not isinstance(value, LACPInterfaces):
            raise TypeError('interfaces must be OpenConfigLACPInterfaces,'
                            'got {}'.format(type(value)))
        self._interfaces = value

    @interfaces.deleter
    def interfaces(self):
        self._interfaces = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('xmlns', self.NAMESPACE)
        if self.interfaces:
            elem.append(self.interfaces.to_xml_element())
        return elem


class LACPInterfaces(abc.Collection):
    """Top-level grouping for LACP-enabled interfaces"""

    NAMESPACE = 'http://openconfig.net/yang/lacp'
    PARENT = 'lacp'
    TAG = 'interfaces'

    def __init__(self):
        # List of interfaces of type OpenconfigInterface
        self._interfaces = list()

    def __iter__(self):
        return iter(self._interfaces)

    def __len__(self):
        return len(self._interfaces)

    def __contains__(self, item):
        return item in self._interfaces

    @property
    def interfaces(self):
        """List of interfaces"""
        return self._interfaces

    def add(self, name: str):
        """Add interface

        :param name: Interface name
        :type: str
        """
        interface = LACPInterface(name)
        self._interfaces.append(interface)
        return interface

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        for interface in self.interfaces:
            elem.append(interface.to_xml_element())
        return elem


class LACPInterface:
    """Base LACP aggregate interface"""

    NAMESPACE = 'http://openconfig.net/yang/lacp'
    PARENT = 'interfaces'
    TAG = 'interface'

    def __init__(self, name: str,
                 operation=constants.NetconfEditConfigOperation.MERGE):
        self.operation = operation
        self._config = LACPInterfaceConfig(name)
        self.name = name

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @property
    def name(self):
        """The name of the LACP aggregate interface."""
        return self._name

    @name.setter
    def name(self, value: str):
        """The name of the LACP aggregate interface."""
        if not isinstance(value, str):
            raise TypeError('name must be string, got {}'.format(type(value)))
        self._name = value
        # 'name' in the configuration is leaf-ref, should match.
        if self.config is not None:
            self.config.name = self.name

    @name.deleter
    def name(self):
        self._name = None

    @property
    def config(self):
        """Configuration data for each LACP aggregate interface"""
        return self._config

    @config.setter
    def config(self, value):
        if not isinstance(value, LACPInterfaceConfig):
            raise TypeError('config must be LACPInterfaceConfig,'
                            'got {}'.format(type(value)))
        self._config = value

    @config.deleter
    def config(self):
        self._config = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('operation', self.operation)
        if self.name:
            common.txt_subelement(elem, 'name', self.name)
        if self.config:
            elem.append(self.config.to_xml_element())
        return elem


class LACPInterfaceConfig:
    """OpenConfig LACP aggregate interface configuration"""

    NAMESPACE = 'http://openconfig.net/yang/lacp'
    PARENT = 'interface'
    TAG = 'config'

    def __init__(self, name: str,
                 operation=constants.NetconfEditConfigOperation.MERGE,
                 interval=types.LACPPeriod.SLOW,
                 lacp_mode=types.LACPActivity.ACTIVE):
        self._name = name
        self.operation = operation
        self.interval = interval
        self.lacp_mode = lacp_mode

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @property
    def name(self):
        """The name of the interface."""
        return self._name

    @name.setter
    def name(self, value: str):
        if not isinstance(value, str):
            raise TypeError('name must be string, got {}'.format(type(value)))
        self._name = value

    @name.deleter
    def name(self):
        self._name = None

    @property
    def interval(self):
        """The period between LACP messages"""
        return self._interval.value if self._interval else None

    @interval.setter
    def interval(self, value: str):
        """Set the period between LACP messages (SLOW or FAST)"""
        if isinstance(value, types.LACPPeriod):
            self._interval = value
        elif isinstance(value, str):
            self._interval = types.LACPPeriod(value)
        else:
            raise TypeError('Invalid type {} for LACP interface interval.'
                            .format(type(value)))

    @interval.deleter
    def interval(self):
        self._interval = None

    @property
    def lacp_mode(self):
        """The LACP mode if the aggregate interface"""
        return self._lacp_mode.value

    @lacp_mode.setter
    def lacp_mode(self, value: str):
        """Set the LACP mode if the aggregate interface

        ACTIVE:  is to initiate the transmission of LACP packets.
        PASSIVE: is to wait for peer to initiate the transmission of
                 LACP packets
        """
        if isinstance(value, types.LACPActivity):
            self._lacp_mode = value
        elif isinstance(value, str):
            self._lacp_mode = types.LACPActivity(value)
        else:
            raise TypeError('Invalid type {} for LACP interface mode.'
                            .format(type(value)))

    @lacp_mode.deleter
    def lacp_mode(self):
        self._lacp_mode = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('operation', self.operation)
        if self.name is not None:
            common.txt_subelement(elem, 'name', self.name)
        if self.interval is not None:
            common.txt_subelement(elem, 'interval', self.interval)
        if self.lacp_mode is not None:
            common.txt_subelement(elem, 'lacp-mode', self.lacp_mode)
        return elem



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from collections import abc
from typing import Optional
from xml.etree import ElementTree

from networking_baremetal import common
from networking_baremetal import constants
from networking_baremetal.openconfig.interfaces import aggregate
from networking_baremetal.openconfig.interfaces import ethernet


class InterfaceConfig:
    """OpenConfig interface configuration"""

    NAMESPACE = 'http://openconfig.net/yang/interfaces'
    PARENT = 'interface'
    TAG = 'config'

    def __init__(self,
                 operation=constants.NetconfEditConfigOperation.MERGE,
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 enabled: Optional[bool] = None,
                 mtu: Optional[int] = None):
        self.operation = operation
        self._name = name
        self._description = None
        self._enabled = None
        self._mtu = None
        if description:
            self.description = description
        if enabled is not None:
            self.enabled = enabled
        if mtu:
            self.mtu = mtu

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value if self._operation else None

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @operation.deleter
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        self._operation = None

    @property
    def name(self):
        """The name of the interface."""
        return self._name

    @name.setter
    def name(self, value: str):
        """The name of the interface."""
        if not isinstance(value, str):
            raise TypeError('name must be string, got {}'.format(type(value)))
        self._name = value

    @name.deleter
    def name(self):
        """The name of the interface."""
        self._name = None

    @property
    def description(self):
        """A textual description of the interface"""
        return self._description

    @description.setter
    def description(self, value: str):
        """A textual description of the interface"""
        if not isinstance(value, str):
            raise TypeError('description must be string, got {}'
                            .format(type(value)))
        self._description = value

    @description.deleter
    def description(self):
        self._description = None

    @property
    def enabled(self):
        """The configured, desired state of the interface"""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        """The configured, desired state of the interface"""
        if not isinstance(value, bool):
            raise TypeError('enabled must be boolean, got {}'
                            .format(type(value)))
        self._enabled = value

    @enabled.deleter
    def enabled(self):
        self._enabled = None

    @property
    def mtu(self):
        """The max transmission unit size in octets"""
        return self._mtu

    @mtu.setter
    def mtu(self, value: int):
        """Set the max transmission unit size in octets"""
        if not isinstance(value, int):
            raise TypeError(f'mtu must be integer, got {type(value)}')
        self._mtu = value

    @mtu.deleter
    def mtu(self):
        self._mtu = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        if self.name is not None:
            common.txt_subelement(elem, 'name', self.name,
                                  attrib={'operation': self.operation})
        if self.description is not None:
            common.txt_subelement(elem, 'description', self.description,
                                  attrib={'operation': self.operation})
        if self.enabled is not None:
            common.txt_subelement(elem, 'enabled', str(self.enabled).lower(),
                                  attrib={'operation': self.operation})
        if self.mtu is not None:
            common.txt_subelement(elem, 'mtu', str(self.mtu),
                                  attrib={'operation': self.operation})
        return elem


class BaseInterface:
    """Base interface"""

    NAMESPACE = 'http://openconfig.net/yang/interfaces'
    PARENT = 'interfaces'
    TAG = 'interface'

    def __init__(self, name: str):
        self.name = name
        self._config = InterfaceConfig()

    @property
    def name(self):
        """The name of the interface."""
        return self._name

    @name.setter
    def name(self, value: str):
        """The name of the interface."""
        if not isinstance(value, str):
            raise TypeError('name must be string, got {}'.format(type(value)))
        self._name = value

    @name.deleter
    def name(self):
        self._name = None

    @property
    def config(self):
        """Configuration parameters for interface"""
        return self._config

    @config.setter
    def config(self, value):
        if not isinstance(value, InterfaceConfig):
            raise TypeError('config must be InterfaceConfig, got {}'
                            .format(type(value)))
        self._config = value

    @config.deleter
    def config(self):
        self._config = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        common.txt_subelement(elem, 'name', self.name)
        if self._config:
            elem.append(self.config.to_xml_element())
        return elem


class InterfaceEthernet(BaseInterface):

    def __init__(self, name: str):
        super(InterfaceEthernet, self).__init__(name)
        self._ethernet = ethernet.InterfacesEthernet()

    @property
    def ethernet(self):
        """Ethernet configuration and state"""
        return self._ethernet

    @ethernet.setter
    def ethernet(self, value):
        if not isinstance(value, ethernet.InterfacesEthernet):
            raise TypeError('ethernet must be InterfacesEthernet, got {}'
                            .format(type(value)))
        self._ethernet = value

    @ethernet.deleter
    def ethernet(self):
        self._ethernet = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        common.txt_subelement(elem, 'name', self.name)
        if self.config:
            elem.append(self.config.to_xml_element())
        if self.ethernet:
            elem.append(self.ethernet.to_xml_element())
        return elem


class InterfaceAggregate(BaseInterface):

    def __init__(self, name: str,
                 operation: str = constants.NetconfEditConfigOperation.MERGE):
        super(InterfaceAggregate, self).__init__(name)
        self.operation = operation
        self._aggregation = aggregate.InterfacesAggregation()

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value if self._operation else None

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @operation.deleter
    def operation(self):
        self._operation = None

    @property
    def aggregation(self):
        """Ethernet configuration and state"""
        return self._aggregation

    @aggregation.setter
    def aggregation(self, value):
        if not isinstance(value, aggregate.InterfacesAggregation):
            raise TypeError('ethernet must be OpenConfigInterfacesAggregation,'
                            'got {}'.format(type(value)))
        self._aggregation = value

    @aggregation.deleter
    def aggregation(self):
        self._aggregation = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        if self.operation:
            elem.set('operation', self.operation)
        common.txt_subelement(elem, 'name', self.name)
        if self.config:
            elem.append(self.config.to_xml_element())
        if self.aggregation:
            elem.append(self.aggregation.to_xml_element())
        return elem


class Interfaces(abc.Collection):
    """Group/List of interfaces"""

    NAMESPACE = 'http://openconfig.net/yang/interfaces'
    TAG = 'interfaces'

    def __init__(self):
        # List of interfaces of type Interface
        self._interfaces = list()

    def __iter__(self):
        return iter(self._interfaces)

    def __len__(self):
        return len(self._interfaces)

    def __contains__(self, item):
        return item in self._interfaces

    @property
    def interfaces(self):
        """List of interfaces"""
        return self._interfaces

    def add(self, name: str,
            interface_type: str = constants.IFACE_TYPE_ETHERNET):
        """Add interface

        :param name: Interface name
        :type: str
        :param interface_type: Interface type ('ethernet', 'aggregate', 'base')
        :type: str
        """
        if interface_type == constants.IFACE_TYPE_ETHERNET:
            interface = InterfaceEthernet(name)
        elif interface_type == constants.IFACE_TYPE_AGGREGATE:
            interface = InterfaceAggregate(name)
        elif interface_type == constants.IFACE_TYPE_BASE:
            interface = BaseInterface(name)
        else:
            raise ValueError('Invalid interface type {}'.format(type(name)))
        self._interfaces.append(interface)
        return interface

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        eleme = ElementTree.Element(self.TAG)
        eleme.set('xmlns', self.NAMESPACE)
        for interface in self.interfaces:
            eleme.append(interface.to_xml_element())
        return eleme



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading
import openstack
from oslo_config import cfg
from oslo_log import log as logging
import tenacity

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

_IRONIC_SESSION = None
IRONIC_GROUP = 'ironic'

_deprecated_opts = {}
_deprecated_opts['endpoint_override'] = [
    cfg.DeprecatedOpt('ironic_url', group=IRONIC_GROUP)]
_deprecated_opts['region_name'] = [
    cfg.DeprecatedOpt('os_region', group=IRONIC_GROUP)]
_deprecated_opts['status_code_retries'] = [
    cfg.DeprecatedOpt('max_retries', group=IRONIC_GROUP)]
_deprecated_opts['status_code_retry_delay'] = [
    cfg.DeprecatedOpt('retry_interval', group=IRONIC_GROUP)]


IRONIC_OPTS = [
    cfg.StrOpt('auth_strategy',
               default='keystone',
               deprecated_for_removal=True,
               deprecated_reason='This option is no longer used, please use '
                                 'the [ironic]/auth_type option instead.',
               choices=('keystone', 'noauth'),
               help='Method to use for authentication: noauth or keystone.'),
]


def list_opts():
    return [
        (IRONIC_GROUP, IRONIC_OPTS
         + loading.get_adapter_conf_options(deprecated_opts=_deprecated_opts)
         + loading.get_session_conf_options(deprecated_opts=_deprecated_opts)
         + loading.get_auth_plugin_conf_options('v3password'))]


def get_session(group):
    loading.register_adapter_conf_options(CONF, group,
                                          deprecated_opts=_deprecated_opts)
    loading.register_session_conf_options(CONF, group,
                                          deprecated_opts=_deprecated_opts)
    loading.register_auth_conf_options(CONF, group)
    CONF.register_opts(IRONIC_OPTS, group=group)
    auth = loading.load_auth_from_conf_options(CONF, group)
    session = loading.load_session_from_conf_options(CONF, group, auth=auth)
    return session


def _get_ironic_session():
    global _IRONIC_SESSION

    if not _IRONIC_SESSION:
        _IRONIC_SESSION = get_session(IRONIC_GROUP)
    return _IRONIC_SESSION


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(openstack.exceptions.NotSupported),
    wait=tenacity.wait_exponential(max=30))
def get_client():
    """Get an ironic client connection."""
    session = _get_ironic_session()

    try:
        return openstack.connection.Connection(
            session=session, oslo_conf=CONF).baremetal
    except openstack.exceptions.NotSupported as exc:
        LOG.error('Ironic API might not be running, failed to establish a '
                  'connection with ironic, reason: %s. Retrying ...', exc)
        raise
    except Exception as exc:
        LOG.error('Failed to establish a connection with ironic, reason: %s',
                  exc)
        raise
